/*
 * $Id: cilk-cilk2c.h,v 1.2 1997/03/01 17:42:38 athena Exp $
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#include <cilk-conf.h>

/* hack : these are constants here, but not in sched.c ! */
DECLARE_PRIVATE_VARIABLE(extern struct ClosureCache *const, cache ,);

#undef offsetof
#define offsetof(type, mem) \
    ((size_t) ((char *)&((type *) 0)->mem - (char *)((type *) 0)))

/* add current frame to deque */
#define _PUSH_FRAME(f) { \
    CILK_ASSERT(_frame->header.magic == CILK_STACKFRAME_MAGIC); \
    Cilk_membar_StoreStore(); \
}

#define _INIT_FRAME(f,s,signat) { \
     volatile StackFrame **_t; \
     f = Cilk_create_frame(s, signat); \
     _t = USE_PRIVATE(cache)->tail; \
     CILK_ASSERT((StackFrame **)_t < USE_PRIVATE(cache)->stack + USE_SHARED_RO(stackdepth)); \
     *_t = (StackFrame *)f; \
     Cilk_membar_StoreStore(); \
     USE_PRIVATE(cache)->tail = _t + 1;\
}

#define _SYNC Cilk_sync()

#define _SYNCHED 0

#define _ABORT_STANDALONE() \
         Cilk_abort_standalone()

#define _ABORT_SLOW() \
         Cilk_abort_slow()

#define _SET_RESULT(x) {\
	 typeof(x) __tmp = x;\
	 Cilk_set_result(&__tmp, sizeof(__tmp));\
}

#define _SET_NORESULT() {\
	 Cilk_set_result((void *)0, 0);\
}

#define _XPOP_FRAME_RESULT(f, r, result) \
{ \
     volatile StackFrame **_t;\
     _t = USE_PRIVATE(cache)->tail; \
     Cilk_membar_StoreLoad(); \
     if (USE_PRIVATE(cache)->exception >= _t) { \
	  typeof(result) __tmp = result; \
	  if (Cilk_exception_handler(&__tmp, sizeof(__tmp))) {\
   	      --USE_PRIVATE(cache)->tail; \
 	      return r; \
          }\
     } \
}

#define _XPOP_FRAME_NORESULT(f, r) \
{ \
     volatile StackFrame **_t;\
     _t = USE_PRIVATE(cache)->tail; \
     Cilk_membar_StoreLoad(); \
     if (USE_PRIVATE(cache)->exception >= _t) { \
 	  if (Cilk_exception_handler((void *)0, 0)) { \
   	     --USE_PRIVATE(cache)->tail; \
	     return r;\
	  }\
     } \
}

#define _BEFORE_SPAWN_FAST() \
     _BEFORE_SPAWN_FAST_CP()

#define _BEFORE_SPAWN_SLOW() \
     _BEFORE_SPAWN_SLOW_CP()

#define _AT_THREAD_BOUNDARY_SLOW() {\
     _AT_THREAD_BOUNDARY_SLOW_CP() \
     WHEN_CILK_STATS(Cilk_event_new_thread();) \
}

#define _AFTER_SPAWN_FAST() {\
     _AFTER_SPAWN_FAST_CP() \
     WHEN_CILK_STATS(Cilk_event_new_thread();) \
}

#define _AFTER_SPAWN_SLOW() \
     _AFTER_SPAWN_SLOW_CP()

#define _START_THREAD_FAST() \
      _START_THREAD_FAST_CP() \
      WHEN_CILK_STATS(Cilk_event_new_thread();)

#define _START_THREAD_SLOW() \
      _START_THREAD_SLOW_CP() \
      CILK_ASSERT(_frame->header.magic == CILK_STACKFRAME_MAGIC); 

#define _BEFORE_RETURN_FAST() \
      _BEFORE_RETURN_FAST_CP() \
      CILK_ASSERT(_frame->header.magic == CILK_STACKFRAME_MAGIC); \
      WHEN_CILK_DEBUG(_frame->header.magic = ~CILK_STACKFRAME_MAGIC); \
      Cilk_destroy_frame((StackFrame *)_frame, sizeof(*_frame)); \
      --USE_PRIVATE(cache)->tail; \


#define _BEFORE_RETURN_SLOW() \
      _BEFORE_RETURN_SLOW_CP() \
      CILK_ASSERT(_frame->header.magic == CILK_STACKFRAME_MAGIC); \
      WHEN_CILK_DEBUG(_frame->header.magic = ~CILK_STACKFRAME_MAGIC); \
      Cilk_destroy_frame((StackFrame *)_frame, sizeof(*_frame)); \
      --USE_PRIVATE(cache)->tail; \

#define _BEFORE_RETURN_INLET()

#define _BEFORE_SYNC_SLOW() \
      _BEFORE_SYNC_SLOW_CP()

#define _AFTER_SYNC_SLOW() \
      _AFTER_SYNC_SLOW_CP() 

#define _AT_SYNC_FAST() {\
      _AT_SYNC_FAST_CP() \
      WHEN_CILK_STATS(Cilk_event_new_thread();) \
}


#if CILK_CRITICAL_PATH
/*
 * Critical-path measurements. Each procedure has two variables,
 * CP and MYCP.
 *
 * Let a "nice node" be the beginning of the thread, or any sync
 * point. CP is less than or equal to MYCP at nice nodes, and MYCP
 * is the actual critical path up to that node.
 *
 * On non-nice nodes, CP and MYCP have different meanings.
 * MYCP is the critical path up to the last nice node, plus the
 * execution time (of the procedure only, not its children) since
 * then. CP is the maximum critical path of all children spawned
 * and finished since last nice node.
 *
 * Therefore, when a child is spawned, it starts with the current
 * value of MYCP; moreover, when sync occurs, CP <= MYCP = max(CP, MYCP).
 *
 * The variable cp_hack is used to pass CP from parent to
 * child, and vice versa.
 *
 * CP and MYCP for the fast version are local variables, saved on the
 * frame before a spawn. The slow version always operates on the frame.
 *
 * We also take advantage of the fact the CP measurements are slow
 * to collect a fine-grained timing of the total work at no appreciable
 * cost. Every thread as a variable WORK that collects the relevant
 * quantities.  The variable work_hack passes this value
 * from child to parent (the vice versa is not needed).
 */


/* at begin of thread, set my critical path to the one of the parent */
#define _START_THREAD_FAST_CP()\
      Cilk_time Cilk_cp = 0, Cilk_mycp = USE_PRIVATE(cp_hack), Cilk_work = 0;\
      WHEN_CILK_USER_TIMERS(USE_PRIVATE(user_work) = Cilk_work);\
      WHEN_CILK_USER_TIMERS(USE_PRIVATE(user_critical_path) = Cilk_mycp);\
      USE_PRIVATE(last_cp_time) = Cilk_get_time();

/* 
 * but the slow version by definition has a valid CP and MYCP, so
 * no initialization is needed.
 */
#define _START_THREAD_SLOW_CP() \
      WHEN_CILK_USER_TIMERS(USE_PRIVATE(user_work) = _frame->header.work);\
      WHEN_CILK_USER_TIMERS(USE_PRIVATE(user_critical_path) = _frame->header.mycp);

#define _AT_THREAD_BOUNDARY_SLOW_CP() \
      USE_PRIVATE(last_cp_time) = Cilk_get_time();

/*
 * before spawn, adjust mycp and pass it to the child (implicitly) 
 */
#define _BEFORE_SPAWN_FAST_CP() {\
      Cilk_time Cilk_elapsed = Cilk_get_elapsed_time();\
      USE_PRIVATE(cp_hack) = (Cilk_mycp += Cilk_elapsed);\
      Cilk_work += Cilk_elapsed;\
      _frame->header.mycp = Cilk_mycp;\
      _frame->header.cp = Cilk_cp;\
      _frame->header.work = Cilk_work;\
}

#define _BEFORE_SPAWN_SLOW_CP() {\
      Cilk_time Cilk_elapsed = Cilk_get_elapsed_time();\
      USE_PRIVATE(cp_hack) = (_frame->header.mycp += Cilk_elapsed);\
      _frame->header.work += Cilk_elapsed;\
}

/* after a spawn, maximize CP and accumulate WORK */
#define _AFTER_SPAWN_FAST_CP()  {\
      if (USE_PRIVATE(cp_hack) > Cilk_cp) Cilk_cp = USE_PRIVATE(cp_hack);\
      Cilk_work += USE_PRIVATE(work_hack);\
      WHEN_CILK_USER_TIMERS(USE_PRIVATE(user_work) = Cilk_work);\
      WHEN_CILK_USER_TIMERS(USE_PRIVATE(user_critical_path) = Cilk_mycp);\
}

#define _AFTER_SPAWN_SLOW_CP()  {\
   if (USE_PRIVATE(cp_hack) > _frame->header.cp) _frame->header.cp = USE_PRIVATE(cp_hack);\
   _frame->header.work += USE_PRIVATE(work_hack);\
   WHEN_CILK_USER_TIMERS(USE_PRIVATE(user_work) = _frame->header.work);\
   WHEN_CILK_USER_TIMERS(USE_PRIVATE(user_critical_path) = _frame->header.mycp);\
}

/* at a sync, set MYCP = CP = max(CP, MYCP) and accumulate work */
#define _AT_SYNC_FAST_CP() {\
      Cilk_time Cilk_elapsed = Cilk_get_elapsed_time();\
      Cilk_mycp += Cilk_elapsed;\
      Cilk_work += Cilk_elapsed;\
      if (Cilk_cp > Cilk_mycp)\
	   Cilk_mycp = Cilk_cp;\
      WHEN_CILK_USER_TIMERS(USE_PRIVATE(user_work) = Cilk_work);\
      WHEN_CILK_USER_TIMERS(USE_PRIVATE(user_critical_path) = Cilk_mycp);\
}

/*
 * the slow case is trickier. Before the check for sync,
 * we update CP/MYCP and work.
 *
 * *After* the check we maximize CP with all CP's of children
 * executed by other processors. We call a special procedure to
 * do that.
 */
#define _BEFORE_SYNC_SLOW_CP() {\
      Cilk_time Cilk_elapsed = Cilk_get_elapsed_time();\
      _frame->header.mycp += Cilk_elapsed;\
      _frame->header.work += Cilk_elapsed;\
      if (_frame->header.cp > _frame->header.mycp)\
	   _frame->header.mycp = _frame->header.cp;\
}

#define _AFTER_SYNC_SLOW_CP() {\
       Cilk_after_sync_slow_cp(&_frame->header.work, &_frame->header.mycp);\
       WHEN_CILK_USER_TIMERS(USE_PRIVATE(user_work) = _frame->header.work);\
       WHEN_CILK_USER_TIMERS(USE_PRIVATE(user_critical_path) = _frame->header.mycp);\
}

/* return is like sync, but also sets CP for the parent */
/* we should be sync'd at a return, so just use mycp */
#define _BEFORE_RETURN_FAST_CP() {\
      Cilk_time Cilk_elapsed = Cilk_get_elapsed_time();\
      USE_PRIVATE(cp_hack) = (Cilk_mycp + Cilk_elapsed);\
      USE_PRIVATE(work_hack) = (Cilk_work + Cilk_elapsed);\
}

#define _BEFORE_RETURN_SLOW_CP() {\
      Cilk_time Cilk_elapsed = Cilk_get_elapsed_time();\
      USE_PRIVATE(cp_hack) = (_frame->header.mycp + Cilk_elapsed);\
      USE_PRIVATE(work_hack) = (_frame->header.work + Cilk_elapsed);\
}

#else

#define _START_THREAD_FAST_CP()
#define _START_THREAD_SLOW_CP()
#define _AT_THREAD_BOUNDARY_SLOW_CP()
#define _BEFORE_SPAWN_FAST_CP()
#define _BEFORE_SPAWN_SLOW_CP()
#define _AFTER_SPAWN_FAST_CP()
#define _AFTER_SPAWN_SLOW_CP()
#define _AT_SYNC_FAST_CP()
#define _BEFORE_SYNC_SLOW_CP()
#define _AFTER_SYNC_SLOW_CP()
#define _BEFORE_RETURN_FAST_CP()
#define _BEFORE_RETURN_SLOW_CP()

#endif
