
/*
 * $Id: stats.c,v 1.2 1997/03/01 17:42:57 athena Exp $
 */

/*
 * $Log: stats.c,v $
 * Revision 1.2  1997/03/01 17:42:57  athena
 * Massive cleanup and merging --- since I was still using the
 * old repository :-(
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

/***********************************************************\
 * Statistics gathering
\***********************************************************/

#include <cilk.h>
#include <cilk-lib.h>
#include <cilk-internal.h>

#if CILK_STATS
#include <stdio.h>

DECLARE_SHARED_RO_VARIABLE(extern FILE *, infofile, );

typedef struct Stats {
     long event_count[EVENT_NTYPES];
     CILK_CACHE_LINE_PAD;
} StatsT;

DECLARE_SHARED_RO_VARIABLE(static struct Stats *, stat_array, );

/*
 * table of description of events. It is printed in
 * the same order.
 */
static const struct event_desc {
     char *name;
     int event;
     int verbosity;   /* if statslevel >= 2 + verbosity this thing is printed */
} event_descriptions[] = {
     {
	  "Threads", EVENT_THREAD,               0
     },
#if 0
     /* this is the same as # of steals, if there are no bugs */
     {
	  "Closures", EVENT_CLOSURE_CREATE,      0
     },
#endif
     {
	  "Steals", EVENT_STEAL,                 0
     },
     {
	  "Attempts", EVENT_STEAL_ATTEMPT,       1
     },
     {
	  "Fail/EmptyQ", EVENT_STEAL_EMPTY_DEQUE,  4
     },
     {
	  "Fail/Dekker", EVENT_STEAL_NO_DEKKER,    4
     },
     {
	  "Fail/Return", EVENT_STEAL_RETURNING,    4
     },
     {
	  "Fail/Abort", EVENT_STEAL_ABORT,         4
     },
     {
	  "Abort/Slow", EVENT_ABORT_SLOW,          0
     },
     {
	  "Abort/Alone", EVENT_ABORT_STANDALONE,   1
     },
     {
	  "Ready->Susp", EVENT_ABORT_READY_SUSPEND,  4
     },
     {
	  "Ready->Ret", EVENT_ABORT_READY_RETURN,    4
     },
     {
	  "Prov. good", EVENT_PROVABLY_GOOD_STEAL,   4
     },
     {
	  "Suspend", EVENT_SUSPEND,                  1
     },
     {
	  "Sync check", EVENT_CILK_SYNC,             1
     },
     {
	  "Exceptions", EVENT_EXCEPTION,             1
     },
     {
	  "Ex/Other", EVENT_EXCEPTION_OTHER,         4
     },
     {
	  "Ex/Steal", EVENT_EXCEPTION_STEAL,         4
     },
     {
	  "Ex/Abort", EVENT_EXCEPTION_ABORT,         4
     },
     {
	  "Ex/Abort_ret", EVENT_EXCEPTION_ABORT_RETURN, 4
     },
     {
	  "Ex/Abort_susp", EVENT_EXCEPTION_ABORT_SUSPEND, 4
     },
     {
	  "Return/slow", EVENT_RETURN_SLOW,          4
     },
     {
	  "Enqueue", EVENT_RETURN_ENQUEUE,           4
     },
     {
	  "Poll", EVENT_POLL_INLETS,                 2
     },
     {
	  "User 0", EVENT_USER0, 0
     },
     {
	  "User 1", EVENT_USER1, 0
     },
     {
	  "User 2", EVENT_USER2, 0
     },
     {
	  "User 3", EVENT_USER3, 0
     },
     {
	  "User 4", EVENT_USER4, 0
     },
     {
	  "User 5", EVENT_USER5, 0
     },
     {
	  "User 6", EVENT_USER7, 0
     },
     {
	  "User 7", EVENT_USER7, 0
     },
     {
	  (char *) 0, 0
     }
};

void Cilk_event_gathering_init(void)
{
     int i, j;

     CILK_ASSERT(Self == -1);
     INIT_SHARED_RO(stat_array,
		    (StatsT *) Cilk_malloc_fixed(sizeof(StatsT) *
						 USE_SHARED_RO(active_size)));

     for (i = 0; i < USE_SHARED_RO(active_size); i++)
	  for (j = 0; j < EVENT_NTYPES; j++)
	       USE_SHARED_RO(stat_array)[i].event_count[j] = 0;
}

#define NAME_DESC "%13s"
#define FIELD_DESC "%9ld"
#define TOTAL_DESC "%10s"
#define TOTAL_FIELD_DESC "%10ld"
#define PROC_PER_LINE 5

DECLARE_SHARED_RO_VARIABLE(extern int, num_threads, );
DECLARE_SHARED_RO_VARIABLE(extern int, num_steals, );

void Cilk_print_rts_statistics(void)
{
     int i, j;
     int start, end, line;
     StatsT total;
     struct event_desc *p;

     CILK_ASSERT(Self == -1);

     for (j = 0; j < EVENT_NTYPES; j++)
	  total.event_count[j] = 0;

     /* compute totals */
     for (i = 0; i < USE_SHARED_RO(active_size); ++i)
	  for (j = 0; j < EVENT_NTYPES; j++)
	       total.event_count[j] += 
		    USE_SHARED_RO(stat_array)[i].event_count[j];

     USE_SHARED_RO(num_threads) = total.event_count[EVENT_THREAD];
     USE_SHARED_RO(num_steals) = total.event_count[EVENT_STEAL] +
                       total.event_count[EVENT_PROVABLY_GOOD_STEAL];

     fprintf(USE_SHARED_RO(infofile),
	     "\nRUNTIME SYSTEM STATISTICS:\n"
	     "\n");

     for (line = 0; 
	  line <= (USE_SHARED_RO(active_size) - 1) / PROC_PER_LINE;
	  ++line) {
	  start = line * PROC_PER_LINE;
	  end = (line + 1) * PROC_PER_LINE;
	  if (end > USE_SHARED_RO(active_size))
	       end = USE_SHARED_RO(active_size);

	  fprintf(USE_SHARED_RO(infofile), NAME_DESC, "PN");
	  for (i = start; i < end; ++i)
	       fprintf(USE_SHARED_RO(infofile), FIELD_DESC, (long) i);

	  if (i == USE_SHARED_RO(active_size))
	       fprintf(USE_SHARED_RO(infofile), TOTAL_DESC, "Total");

	  fprintf(USE_SHARED_RO(infofile), "\n");

	  for (p = (struct event_desc *) event_descriptions; p->name; p++) {
	       if (!total.event_count[p->event])
		    continue;  /* skip if total is 0 */

	       if (p->verbosity > USE_SHARED_RO(StatLevel) - 2)
		    continue;  /* skip if stat level too low */

	       fprintf(USE_SHARED_RO(infofile), NAME_DESC, p->name);
	       for (i = start; i < end; ++i)
		    fprintf(USE_SHARED_RO(infofile), FIELD_DESC,
			    USE_SHARED_RO(stat_array)[i].event_count[p->event]);
	       if (i == USE_SHARED_RO(active_size))
		    fprintf(USE_SHARED_RO(infofile), TOTAL_FIELD_DESC,
			    total.event_count[p->event]);
	       fprintf(USE_SHARED_RO(infofile), "\n");
	  }

	  fprintf(USE_SHARED_RO(infofile), "\n");

     }
}

void Cilk_event(int type)
{
     if (Self != -1)
	  ++USE_SHARED_RO(stat_array)[Self].event_count[type];
}

void Cilk_event_new_thread(void)
{
     if (Self != -1)
	  ++USE_SHARED_RO(stat_array)[Self].event_count[EVENT_THREAD];
}

#endif				/* CILK_STATS */
